import java.util.concurrent.Semaphore; // Semaphores library for Java

class ProducerConsumer {
	final int CAPACITY = 10;
	final int CONSUMERS = 1;
	final int PRODUCERS = 1;
	
    private int write;
    private int read;
	
    int[] buffer = new int[CAPACITY];
    
    Semaphore empty = new Semaphore(CAPACITY);
    Semaphore full = new Semaphore(0);
    
    public ProducerConsumer() {
    	// Allocate the threads and start them
    	for (int i=0 ; i<CONSUMERS ; ++i) {
    		Thread t = new Thread(new Consumer(i));
    		t.start();
    	}
    	for (int i=0 ; i<PRODUCERS ; ++i) {
    		Thread t = new Thread(new Producer(i));
    		t.start();
    	}
    	
    }

    public static void main(String[] args) {
        new ProducerConsumer();
    }

    class Producer implements Runnable {        
        private int id;
        
        Producer(int i) {
        	id = i;
        }
    	
        public void run() {
        	while (true) {
        		try {
        			boolean ack = false;
        			while (!ack) {
        				synchronized (empty) {
        					if (empty.tryAcquire()) {
        						ack = true;
        					} else {
        						empty.wait();
        					}
        				}
        			}
        			
					buffer[write] = 1;
					System.out.println("Produced " + write + " (" + id + ")");
					write = (write + 1)%CAPACITY;

					synchronized (full) {
						full.release();
						full.notifyAll();
					}
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
    }

    class Consumer implements Runnable {
        private int id;
        
        Consumer(int i) {
        	id = i;
        }

        public void run() {
        	while (true) {
        		try {
        			boolean ack = false;
        			while (!ack) {
        				synchronized (full) {
        					if (full.tryAcquire()) {
        						ack = true;
        					} else {
        						// take some action if needed
        						full.wait();
        					}
        				}
        			}

        			buffer[read] = 0;
					System.out.println("\t\t\t\t\tConsumed " + read + " (" + id + ")");
					read = (read + 1) %CAPACITY;

					synchronized (empty) {
						empty.release();
						empty.notifyAll();
					}
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
    }



}
