import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

// 1. Create a counter shared between two threads and increase it
// 3. Add a lock externally and protect accesses
// 2. Discuss the difference between implements and extends
// 4. Replace the lock by synchronization (on object)
// 5. Replace the lock by binary semaphore (redo for sharing)
// 6. Protect the counter by adding built in synchronization
// 7. Try combinations of different locking
// 7. Implement producer consumer system using semaphore
// 8. Show the barrier problem and its solution


class SharedCounterWInternalSync {

	final int target;
	final int numthreads;
	final Thread[] threads;
	
	// Count how many people entered overall. 
	final Counter counter;
	
	Semaphore sem = new Semaphore(1);
	
	// Here we start the program
	public static void main(String[] args) {
		SharedCounterWInternalSync c = new SharedCounterWInternalSync(10,10000);

		long startTime = System.currentTimeMillis();
		c.count();
		long stopTime = System.currentTimeMillis();
		long elapsedTime = stopTime-startTime;

		// Print the final result of `counter`
		System.out.println("counter = " + c.value());
		System.out.println("elapsed time = " + elapsedTime);
		System.exit(1);
	}
	

	public SharedCounterWInternalSync(int numthreads, int target) {
		this.target = target;
		this.numthreads = numthreads;

		counter = new Counter(0); 
		
		threads = new Thread[numthreads];
		for (int i=0 ; i<numthreads ; ++i) {
			threads[i] = new Thread(new Increaser());
		}
	}
	
	public void count() {

		try {
			for (int i=0 ; i<numthreads ; ++i) {
				threads[i].start();
			}

			for (int i=0 ; i<numthreads ; ++i) {
				threads[i].join();
			}
		} catch (InterruptedException e) { // Thread methods such as `start()` or `join()` may through `InterruptedException`
			System.out.println(e);
		}
	}

	public int value() {
		return counter.value();
	}
	
	class Increaser implements Runnable {

		// The `run()` method is executed when calling `start()`
		public void run() {
			for (int i = 0; i < target; i++) {
				counter.inc();
			}
		}
	}
	
	class Counter {
		private int value;

		public Counter(int v) {
			value = v;
		}

		public synchronized void inc() {
			// value = value + 1;
			++value;
		}
		
		public synchronized void do_something() {
			// do something
			// This method is also executed separately from inc
		}
		
		public int value() {
			return value;
		}
	}

}
