import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

// 1. Create a counter shared between two threads and increase it
// 3. Add a lock externally and protect accesses
// 2. Discuss the difference between implements and extends
// 4. Replace the lock by synchronization (on object)
// 5. Replace the lock by binary semaphore
// 6. Protect the counter by adding built in synchronization
// 7. Try combinations of different locking
// 7. Implement producer consumer system using semaphore
// 8. Show the barrier problem and its solution


class SharedCounterWSemaphore {

	final int target;
	final int numthreads;
	final Thread[] threads;
	
	Semaphore sem = new Semaphore(1);
	
	// Count how many people entered overall. 
	final Counter counter;
	
	// Here we start the program
	public static void main(String[] args) {
		SharedCounterWSemaphore c = new SharedCounterWSemaphore(10,10000);

		long startTime = System.currentTimeMillis();
		c.count();
		long stopTime = System.currentTimeMillis();
		long elapsedTime = stopTime-startTime;

		// Print the final result of `counter`
		System.out.println("counter = " + c.value());
		System.out.println("elapsed time = " + elapsedTime);
		System.exit(1);
	}
	

	public SharedCounterWSemaphore(int numthreads, int target) {
		this.target = target;
		this.numthreads = numthreads;

		counter = new Counter(0); 
		
		threads = new Thread[numthreads];
		for (int i=0 ; i<numthreads ; ++i) {
			threads[i] = new Increaser();
		}
	}
	
	public void count() {

		try {
			for (int i=0 ; i<numthreads ; ++i) {
				threads[i].start();
			}

			for (int i=0 ; i<numthreads ; ++i) {
				threads[i].join();
			}
		} catch (InterruptedException e) { // Thread methods such as `start()` or `join()` may through `InterruptedException`
			System.out.println(e);
		}
	}

	public int value() {
		return counter.value();
	}
	
	class Increaser extends Thread {

		// The `run()` method is executed when calling `start()`
		public void run() {
			for (int i = 0; i < target; i++) {
				try {
					sem.acquire();
					counter.inc();
					sem.release();
				} catch (InterruptedException e) {
					// This code is run because thread
					// tried to acquire went to sleep and
					// woken up by interrupt.
					// This means that the thread DIDN'T
					// do what it was supposed to do (count)
					// So reduce i by 1 and the loop will run one
					// more time.
					--i; 
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}
	class Counter {
		private int value;

		public Counter(int v) {
			value = v;
		}

		public void inc() {
			// value = value + 1;
			++value;
		}
		
		public int value() {
			return value;
		}
	}

}
